r"""
**This module allows you to record the state of your neural network in various situations
(for example activations value passing through it).**

Recorders are organized similarly to `torch.nn.Module`'s hooks (e.g. `backward`, `forward` and `forward pre`).
Those will record data going out/in from/to `torch.nn.Module` in backward or forward mode.

For example::

    # MNIST classifier
    model = torch.nn.Sequential(
        torch.nn.Linear(784, 100),
        torch.nn.ReLU(),
        torch.nn.Linear(100, 50),
        torch.nn.ReLU(),
        torch.nn.Linear(50, 10),
    )
    # Recorder which sums layer inputs from consecutive forward calls
    recorder = torchfunc.record.ForwardPreRecorder(reduction=lambda x, y: x+y)
    # Record inputs going into Linear(100, 50) and Linear(50, 10)
    recorder.children(model, indices=(2, 3))
    # Train your network normally (or pass data through it)
    ...
    # Save tensors (of shape 100 and 50) in folder, each named 1.pt and 2.pt respectively
    recorder.save(pathlib.Path("./analysis"))

You could specify types instead of indices (all inputs to torch.nn.Linear will be registered),
iterate over modules recursively instead of children etc.

See exact specification below.


"""

import dataclasses
import inspect
import pathlib
import typing

import torch

from ._base import Base


class _Recorder(Base):
    r"""**{}**

    Parameters
    ----------
    condition : Callable, optional
        No argument callable. If True returned, record data.
        Can be used to save data based on external environment (e.g. dataset's label).
        If None, will record every data point. Default: `None`
    reduction : Callable, optional
        Operation to use on incoming data. Should take two arguments, and return one.
        Acts similarly to reduction argument of Python's itertools.reduce.
        If None, data will be added to list, which may be very memory intensive.
        Default: `None`

    Attributes
    ----------
    data : List
        Keeps data passing through hooks, optionally reduced by `reduction`.
        Each item represents data for specified layer
        (with the lower indices being closer to torch.nn.Module's input).
    handles : List[torch.utils.hooks.RemovableHandle]
        Handles for registered hooks, each corresponds to specific submodule.
        Can be used to unregister certain hooks (though discouraged).
    samples : int
        How many samples passed the condition. Useful for calculating running operations
        (e.g. running mean).

    """

    def __init__(self, register_method, method):
        self._register_method: typing.Callable = register_method
        self._method: typing.Callable = method
        self.data = []
        self.hooks = []
        self.handles = []

    def _register_hook(
        self,
        network,
        iterating_function: str,
        types: typing.Tuple[typing.Any] = None,
        indices: typing.List[int] = None,
    ):
        last_index = 0
        for index, module in enumerate(getattr(network, iterating_function)()):
            if isinstance(module, types) or (indices is not None and index in indices):
                hook = self._method(last_index, self.data)
                self.handles.append(getattr(module, self._register_method)(hook))
                self.hooks.append(hook)
                last_index += 1

    def handle(self, index):
        return self.handles[index]

    def samples(self, index):
        return self.hooks[index].samples

    def iter_samples(self):
        for hook in self.hooks:
            yield hook.samples

    def modules(
        self,
        module,
        types: typing.Tuple[typing.Any] = None,
        indices: typing.List[int] = None,
    ):
        r"""**Recursively register gathering hook for specified module, types and/or indices.**

        This reduction will use `modules` method of `torch.nn.Module` to iterate over available submodules.
        If you wish to iterate non-recursively, use `children`.

        Parameters
        ----------
        module : torch.nn.Module
            Module (usually neural network) for which inputs will be collected.
        types : Tuple[typing.Any], optional
            Types whose input's will be recorded. E.g. [torch.nn.Conv2d, torch.nn.Linear] will register
            hooks on every module module of those types. By default, no hook will be registered based on type.
            If you want to register input to every module/child, use object base class.
            Default: `None`
        indices : Iterable[int], optional
            Indices of modules whose inputs will be registered.
            Default: `None` (no modules will have hooks registered based on index).

        Returns
        -------
        self

        """
        self._register_hook(module, "modules", types, indices)
        return self

    def children(
        self,
        network,
        types: typing.Tuple[typing.Any] = None,
        indices: typing.List[int] = None,
    ):

        r"""**Non-recursively register gathering hook for specified module, types and/or indices.**

        This reduction will use `children` method of `torch.nn.Module` to iterate over available submodules.
        This is a shallow iteration (only submodules of current module).

        Parameters
        ----------
        network: torch.nn.Module
                Module (usually neural network) for which inputs will be collected.
        types: typing.Tuple[typing.Any], optional
                Types whose input's will be recorded. E.g. [torch.nn.Conv2d, torch.nn.Linear] will register
                hooks on every module module of this type. By default, no hook will be registered based on type.
                If you want to register input to every module/child, use object base class.
                Default: None
        indices: typing.Iterable[int], optional
                Indices of modules whose inputs will be registered.
                Default: None (no modules will have hooks registered based on index).

        Returns
        -------
        self

        """
        self._register_hook(network, "children", types, indices)
        return self

    def save(self, path: pathlib.Path, mkdir: bool = False, *args, **kwargs):
        """Save data tensors to specified path.

        Each data tensor will be indexed by integer `[0, N)`, where those with smaller
        indices are closer to the input. Ordering is the same as in `__getitem__` or `__iter__`.

        Parameters
        ----------
        path: pathlib.Path
                Path where tensors will be saved.
        mkdir: bool, optional
                If True, create directory if doesn't exists. Default: False
        *args:
                Varargs passed to pathlib.Path's mkdir method if mkdir argument set to True.
        *kwargs:
                Kwarargs passed to pathlib.Path's mkdir method if mkdir argument set to True.

        """
        if mkdir:
            path.mkdir(*args, **kwargs)
        for index, module in enumerate(self):
            torch.save(module, path / f"{index}.pt")

    def __setitem__(self, index, item):
        self.data[index] = item

    def __getitem__(self, index):
        return self.data[index]

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)


@dataclasses.dataclass
class _Hook:
    index: int
    data: typing.List
    samples: int = 0

    def _call(self, to_record, condition, reduction):
        if condition() or condition is None:
            self.samples += 1
            if self.index >= len(self.data):
                self.data.append(to_record[0])
                if reduction is None:
                    self.data[-1] = [self.data[-1]]
            else:
                if reduction is not None:
                    self.data[self.index] = reduction(
                        self.data[self.index], to_record[0]
                    )
                else:
                    self.data[self.index].append(to_record[0])


@dataclasses.dataclass(repr=False)
class ForwardPreRecorder(_Recorder):
    __doc__ = _Recorder.__doc__.format(
        "Record input values before forward of specified layer(s)."
    )

    condition: typing.Callable = None
    reduction: typing.Callable = None

    def __post_init__(self):
        class ForwardPreHook(_Hook):
            def __call__(inner_self, module, inputs):
                inner_self._call(inputs, self.condition, self.reduction)

        super().__init__("register_forward_pre_hook", ForwardPreHook)


class ForwardInputRecorder(_Recorder):
    __doc__ = _Recorder.__doc__.format(
        "Record input values after forward of specified layer(s)."
    )

    condition: typing.Callable = None
    reduction: typing.Callable = None

    def __post_init__(self):
        class ForwardInputHook(_Hook):
            def __call__(inner_self, module, inputs, _):
                inner_self._call(inputs, self.condition, self.reduction)

        super().__init__("register_forward_hook", ForwardInputHook)


class ForwardOutputRecorder(_Recorder):
    __doc__ = _Recorder.__doc__.format(
        "Record output values after forward of specified layer(s)."
    )

    condition: typing.Callable = None
    reduction: typing.Callable = None

    def __post_init__(self):
        class ForwardOutputHook(_Hook):
            def __call__(inner_self, module, _, outputs):
                inner_self._call(outputs, self.condition, self.reduction)

        super().__init__("register_forward_hook", ForwardOutputHook)


class BackwardInputRecorder(_Recorder):
    __doc__ = _Recorder.__doc__.format(
        "Record input gradients after those are calculated w.r.t. specified module."
    )

    condition: typing.Callable = None
    reduction: typing.Callable = None

    def __post_init__(self):
        class BackwardInputHook(_Hook):
            def __call__(inner_self, module, grad_inputs, _):
                inner_self._call(grad_inputs, self.condition, self.reduction)

        super().__init__("register_backward_hook", BackwardInputHook)


class BackwardOutputRecorder(_Recorder):
    __doc__ = _Recorder.__doc__.format(
        "Record output gradients after those are calculated w.r.t. specified module."
    )

    condition: typing.Callable = None
    reduction: typing.Callable = None

    def __post_init__(self):
        class BackwardOutputHook(_Hook):
            def __call__(inner_self, module, _, outputs):
                inner_self._call(outputs, self.condition, self.reduction)

        super().__init__("register_backward_hook", BackwardOutputHook)



r"""
**This module allows you to plot and visualize specific parts of your neural network.**

Currently you can display linear activations and their strength using `Activations1d`.
For future plans on visualization see `ROADMAP` document in GitHub repository.

"""

import builtins
import dataclasses
import typing

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import torch

from ._base import Base


@dataclasses.dataclass(repr=False)
class Activations1d(Base):
    """**Plot activations given as 1D tensors.**

    One can easily visualize activations of torch.nn.Linear module
    (or any other returning 1d Tensor as output from the layer)
    Activations can be recorded with record module.

    Can be used in conjunction with `torchfunc.record` module.

    Example::

        import matplotlib.pyplot as plt

        ... # Define recorder previously
        fig = plt.figure(figsize=(12, 12))
        ax = fig.gca()

        plotter = torchfunc.plot.Activations1d(ax=ax)
        # Pass 1d tensors, here from recorder
        plotter(*[data for data in recorder]) # See recorder module
        fig.savefig(f"network_activations.png")

    Attributes
    ----------
    ax : matplotlib.axis.Axis
        Matplotlib axis where activations will be plotted
    colormap : matplotlib.colors.Color, optional
        Instance of colormap. See https://matplotlib.org/3.1.1/gallery/color/colormap_reference.html.
        You can use `matplotlib.pyplot.get_cmap(name)` to get cmap.
        Default: `None` (default colormap)
    horizontal : bool, optional
        Whether network should be drawn horizontally. Default: False
    value_max : int, optional
        Maximum value of neuron activation that will be displayed on colorbar.
        If not specified, it will be taken from data passed via __call__.
    value_min : int, optional
        Minimum value of neuron activation that will be displayed on colorbar.
        If not specified, it will be taken from data passed via __call__.
        Default: None
    neuron_size : int, optional
        Size of single neuron. It is the same as size argument of `matplotlib.pyplot.scatter`'s `s` (size) argument.
        Default: None
    edge_colors : typing.Any, optional
        Edge colors of neurons. It is the same as `edge_colors` argument of `matplotlib.pyplot.scatter`.
        Default: None
    orientation : str, optional
        Location of colorbar. Same as `matplotlib.pyplot.colorbar`'s orientation argument.
        Default: vertical

    """

    ax: matplotlib.axis.Axis
    colormap: typing.Any = None
    horizontal: bool = False
    value_max: int = None
    value_min: int = None
    neuron_size: int = None
    edge_colors: typing.Any = None
    orientation: str = "vertical"

    def __post_init__(self):
        self.ax.axis("off")

    def __call__(self, *activations: torch.Tensor) -> None:
        def _find_extremes(activations, initial_value, function: str):
            return (
                getattr(builtins, function)(map(getattr(torch, function), activations))
                if initial_value is None
                else initial_value
            )

        if activations:
            vmin, vmax = (
                _find_extremes(activations, self.value_min, "min"),
                _find_extremes(activations, self.value_max, "max"),
            )
            longest_layer = max(map(len, activations))

            last_scatter = None
            for index, layer in enumerate(activations):
                spacing = int((longest_layer - len(layer)) * 0.5)
                layer_position_in_model = np.full(len(layer), index)
                neuron_positions = range(spacing, spacing + len(layer))
                last_scatter = self.ax.scatter(
                    x=neuron_positions if self.horizontal else layer_position_in_model,
                    y=layer_position_in_model if self.horizontal else neuron_positions,
                    c=layer.float().numpy(),
                    cmap=self.colormap,
                    s=self.neuron_size,
                    vmin=vmin,
                    vmax=vmax,
                    edgecolors=self.edge_colors,
                )

            plt.colorbar(last_scatter, ax=self.ax, orientation=self.orientation)



r"""
**This module provides functionalities related to torch.nn.Module instances (e.g. freezing parameters).**

For performance analysis of `torch.nn.Module` please see subpackage `performance`.

"""

import copy
import datetime
import pathlib

import torch

from ._base import Base


def _switch(module: torch.nn.Module, weight: bool, bias: bool, value: bool):
    for name, param in module.named_parameters():
        if bias and weight:
            param.requires_grad_(value)
        elif ("bias" in name and bias) or ("weight" in name and weight):
            param.requires_grad_(value)

    return module


def freeze(
    module: torch.nn.Module, weight: bool = True, bias: bool = True
) -> torch.nn.Module:
    r"""**Freeze module's parameters.**

    Sets `requires_grad` to `False` for specified parameters in module.
    If bias and weight are specified, ALL parameters will be frozen (even if their names
    are not matched by `weight` and bias).

    If you want to freeze only those whose names contain `bias` or `weight`,
    call the function twice consecutively (once with `bias=True` and `weight=False` and vice versa).

    Example::

        logistic_regression = torch.nn.Sequential(
            torch.nn.Linear(784, 10),
            torch.nn.Sigmoid(),
        )

        # Freeze only bias in logistic regression
        torchfunc.freeze(logistic_regression, weight = False)


    Parameters
    ----------
    module : torch.nn.Module
            Module whose weights and biases will be frozen.
    weight : bool, optional
            Freeze weights. Default: True
    bias : bool, optional
            Freeze bias. Default: True

    Returns
    -------
    module : torch.nn.Module
        Module after parameters were frozen

    """
    return _switch(module, weight, bias, value=False)


def unfreeze(
    module: torch.nn.Module, weight: bool = True, bias: bool = True
) -> torch.nn.Module:
    r"""**Unfreeze module's parameters.**

    Sets `requires_grad` to `True` for all parameters in module.
    Works as complementary function to freeze, see it's documentation.

    Parameters
    ----------
    module : torch.nn.Module
            Module whose weights and biases will be unfrozen.
    weight : bool, optional
            Freeze weights. Default: True
    bias : bool, optional
            Freeze bias. Default: True

    Returns
    -------
    module : torch.nn.Module
        Module after parameters were unfrozen

    """
    return _switch(module, weight, bias, value=False)


class Snapshot(Base):
    r"""**Save module snapshots in memory and/or disk.**

    Next models can be added with `+` or `+=` and their state or whole model saved
    to disk with appropriate methods.

    All added modules are saved unless removed with `pop()` method.

    Example::

        snapshot = torchfunc.module.Snapshot()
        snapshot += torch.nn.Sequential(torch.nn.Linear(784, 10), torch.nn.Sigmoid())
        snapshot.save() # Saved last model to disk with default settings

    Parameters
    ----------
    *modules : torch.nn.Module
        Var args of PyTorch modules to be kept.

    """

    def __init__(self, *modules: torch.nn.Module):
        if modules:
            self.modules = list(modules)
        self.modules = []
        self.timestamps = []

    def __iadd__(self, other):
        self.modules.append(other)
        self.timestamps.append(datetime.datetime.now())
        return self

    def __radd__(self, other):
        return self + other

    def __add__(self, other: torch.nn.Module):
        new = Snapshot(copy.deepcopy(self.modules))
        new += other
        return new

    def pop(self, index: int = -1):
        r"""**Remove module at `index` from memory.**


        Parameters
        ----------
        index : int, optional
            Index of module to be removed. Default: -1 (last module)

        Returns
        ----------
        module : torch.nn.Module
            Module removed by this operation

        """
        return self.modules.pop(index)

    def _save(
        self, index, folder: pathlib.Path, name: pathlib.Path, remove: bool, state: bool
    ) -> None:
        module = self.modules[index]
        if state:
            module = module.state_dict()
        if folder is None:
            folder = pathlib.Path(".")
        if name is None:
            name = pathlib.Path(
                f"module_" + "state_"
                if state
                else "" + f"{len(self.modules)}_{self.timestamps[index]}.pt"
            )
        if remove:
            self.pop(index)
        torch.save(module, folder / name)

    def save(
        self,
        folder: pathlib.Path = None,
        name: pathlib.Path = None,
        index: int = -1,
        remove: bool = False,
    ) -> None:
        r"""**Save module to disk.**

        See PyTorch's docs for more information: https://pytorch.org/tutorials/beginner/saving_loading_models.html#save-load-entire-model

        Parameters
        ----------
        folder : pathlib.Path, optional
            Name of the folder where model will be saved. Defaults to current working directory.
        name : pathlib.Path, optional
            Name of the file. Default: module_{index}_{timestamp}.pt
        index : int, optional
            Index of the module to be saved. Default: Last module
        remove : bool, optional
            Whether module should be removed from memory after saving. Useful
            for keeping only best/last model in memory. Default: False

        """
        self._save(index, folder, name, remove, state=False)

    def save_state(
        self,
        folder: pathlib.Path = None,
        name: pathlib.Path = None,
        index: int = -1,
        remove: bool = False,
    ) -> None:
        r"""**Save module's state to disk.**

        See PyTorch's docs on `state_dict` for more information: https://pytorch.org/tutorials/beginner/saving_loading_models.html#save-load-state-dict-recommended

        Parameters
        ----------
        folder : pathlib.Path, optional
            Name of the folder where model will be saved. Defaults to current working directory.
        name : pathlib.Path, optional
            Name of the file. Default: module_{index}_{timestamp}.pt
        index : int, optional
            Index of the module to be saved. Default: Last module
        remove : bool, optional
            Whether module should be removed from memory after saving. Useful
            for keeping only best/last model in memory. Default: False

        """
        self._save(index, folder, name, remove, state=True)



import itertools

import torch

import torchfunc


def _test_value(condition, parameter):
    if condition:
        return not parameter.requires_grad
    return parameter.requires_grad


def _test_network(model: bool, bias: bool, weights: bool):
    module = (
        torch.nn.Sequential(
            torch.nn.Linear(100, 50), torch.nn.Linear(50, 50), torch.nn.Linear(50, 10)
        )
        if model
        else torch.nn.Linear(20, 40)
    )

    for parameter in module.parameters():
        assert parameter.requires_grad
    for name, parameter in torchfunc.module.freeze(
        module, weights, bias
    ).named_parameters():
        print(name)
        assert _test_value(bias if "bias" in name else weights, parameter)


def test_freezing():
    for model, bias, weights in itertools.product([True, False], repeat=3):
        print(model, bias, weights)
        _test_network(model, bias, weights)
